@@ -320,7 +320,7 @@ namespace {
 	public:
 		const static int N_FILES = 8;
 		const static int N_SETS = 256;
-#ifdef DEV_BUILD
+#if defined(DEV_BUILD)
 		const static int SEM_PER_SET = 4;	// force multiple sets allocation
 #else
 		const static int SEM_PER_SET = 31;	// hard limit for some old systems, might set to 32
@@ -635,6 +635,7 @@ namespace {
 					{
 						semctl(id, 0, IPC_RMID);
 					}
+					set[n].fileNum = -1;
 				}
 				idCache[n] = -1;
 			}
@@ -2090,6 +2091,16 @@ UCHAR* ISC_map_file(ISC_STATUS* status_vector,
 		return NULL;
 	}
 
+	if (!init_flag)
+	{
+		if ((GetLastError() != ERROR_ALREADY_EXISTS) ||	SetFilePointer(file_handle, 0, NULL, FILE_END) == 0)
+		{
+			CloseHandle(event_handle);
+			CloseHandle(file_handle);
+			goto retry;
+		}
+	}
+
 /* Create a file mapping object that will be used to make remapping possible.
    The current length of real mapped file and its name are saved in it. */
 
@@ -2212,12 +2223,19 @@ UCHAR* ISC_map_file(ISC_STATUS* status_vector,
 	if (init_flag)
 	{
 		FlushViewOfFile(address, 0);
-		SetEvent(event_handle);
+
+		DWORD err = 0;
 		if (SetFilePointer(shmem_data->sh_mem_handle, length, NULL, FILE_BEGIN) == INVALID_SET_FILE_POINTER ||
 			!SetEndOfFile(shmem_data->sh_mem_handle) ||
 			!FlushViewOfFile(shmem_data->sh_mem_address, 0))
 		{
-			error(status_vector, "SetFilePointer", GetLastError());
+			err = GetLastError();
+		}
+
+		SetEvent(event_handle);
+		if (err)
+		{
+			error(status_vector, "SetFilePointer", err);
 			return NULL;
 		}
 	}
@@ -2796,6 +2814,7 @@ static const LPCSTR FAST_MUTEX_MAP_NAME	= "%s_FM_MAP";
 static const int DEFAULT_INTERLOCKED_SPIN_COUNT	= 0;
 static const int DEFAULT_INTERLOCKED_SPIN_COUNT_SMP	= 200;
 
+static SLONG pid = 0;
 
 typedef WINBASEAPI BOOL (WINAPI *pfnSwitchToThread) ();
 static inline BOOL switchToThread()
@@ -2887,6 +2906,7 @@ static DWORD enterFastMutex(FAST_MUTEX* lpMutex, DWORD dwMilliseconds)
 #ifdef _DEBUG
 			lpSect->dwThreadId = GetCurrentThreadId();
 #endif
+			lpSect->lOwnerPID = pid;
 			unlockSharedSection(lpSect);
 			return WAIT_OBJECT_0;
 		}
@@ -2905,11 +2925,37 @@ static DWORD enterFastMutex(FAST_MUTEX* lpMutex, DWORD dwMilliseconds)
 		unlockSharedSection(lpSect);
 
 		// TODO actual timeout can be of any length
-		const DWORD dwResult = WaitForSingleObject(lpMutex->hEvent, dwMilliseconds);
+		const DWORD tm = (dwMilliseconds == INFINITE || dwMilliseconds > 5000) ? 5000 :  dwMilliseconds;
+		const DWORD dwResult = WaitForSingleObject(lpMutex->hEvent, tm);
+
 		InterlockedDecrement(FIX_TYPE(&lpSect->lThreadsWaiting));
 
-		if (dwResult != WAIT_OBJECT_0)
+		if (dwMilliseconds != INFINITE)
+			dwMilliseconds -= tm;
+
+//		if (dwResult != WAIT_OBJECT_0)
+//			return dwResult;
+
+		if (dwResult == WAIT_OBJECT_0)
+			continue;
+		if (dwResult == WAIT_ABANDONED)
 			return dwResult;
+		if (dwResult == WAIT_TIMEOUT && !dwMilliseconds)
+			return dwResult;
+
+		lockSharedSection(lpSect, lpMutex->lSpinCount);
+		if (lpSect->lOwnerPID > 0 && !lpSect->lAvailable)
+		{
+			if (!ISC_check_process_existence(lpSect->lOwnerPID))
+			{
+#ifdef DEBUG
+				gds__log("enterFastMutex: dead process detected, pid = %d", lpSect->lOwnerPID);
+#endif
+				lpSect->lOwnerPID = 0;
+				lpSect->lAvailable++;
+			}
+		}
+		unlockSharedSection(lpSect);
 	}
 }
 
@@ -2927,6 +2973,7 @@ static bool leaveFastMutex(FAST_MUTEX* lpMutex)
 	lpSect->lAvailable++;
 	if (lpSect->lThreadsWaiting)
 		SetEvent(lpMutex->hEvent);
+	lpSect->lOwnerPID = -pid;
 	unlockSharedSection(lpSect);
 
 	return true;
@@ -2953,6 +3000,9 @@ static inline void setupMutex(FAST_MUTEX* lpMutex)
 static bool initializeFastMutex(FAST_MUTEX* lpMutex, LPSECURITY_ATTRIBUTES lpAttributes,
 								BOOL bInitialState, LPCSTR lpName)
 {
+	if (pid == 0)
+		pid = GetCurrentProcessId();
+
 	LPCSTR name = lpName;
 
 	if (strlen(lpName) + strlen(FAST_MUTEX_EVT_NAME) - 2 >= MAXPATHLEN)
@@ -2971,6 +3021,11 @@ static bool initializeFastMutex(FAST_MUTEX* lpMutex, LPSECURITY_ATTRIBUTES lpAtt
 		name = sz;
 	}
 
+#ifdef DONT_USE_FAST_MUTEX
+	lpMutex->lpSharedInfo = NULL;
+	lpMutex->hEvent = CreateMutex(lpAttributes, bInitialState, name);
+	return (lpMutex->hEvent != NULL);
+#else
 	lpMutex->hEvent = CreateEvent(lpAttributes, FALSE, FALSE, name);
 	DWORD dwLastError = GetLastError();
 
@@ -3001,6 +3056,7 @@ static bool initializeFastMutex(FAST_MUTEX* lpMutex, LPSECURITY_ATTRIBUTES lpAtt
 					lpMutex->lpSharedInfo->lSpinLock = 0;
 					lpMutex->lpSharedInfo->lThreadsWaiting = 0;
 					lpMutex->lpSharedInfo->lAvailable = bInitialState ? 0 : 1;
+					lpMutex->lpSharedInfo->lOwnerPID = bInitialState ? pid : 0;
 					InterlockedExchange(FIX_TYPE(&lpMutex->lpSharedInfo->fInitialized), 1);
 				}
 				else
@@ -3019,6 +3075,7 @@ static bool initializeFastMutex(FAST_MUTEX* lpMutex, LPSECURITY_ATTRIBUTES lpAtt
 
 	SetLastError(dwLastError);
 	return false;
+#endif // DONT_USE_FAST_MUTEX
 }
 
 #ifdef NOT_USED_OR_REPLACED
